# Adapted from https://github.com/encode/broadcaster

import asyncio
import typing
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, AsyncIterator, Dict


class Event:
    def __init__(self, channel: str, message: typing.Any) -> None:
        self.channel = channel
        self.message = message

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Event)
            and self.channel == other.channel
            and self.message == other.message
        )

    def __repr__(self) -> str:
        return f"Event(channel={self.channel!r}, message={self.message!r})"


class Subscriber:
    def __init__(self, queue: asyncio.Queue[Event | None]) -> None:
        self._queue = queue

    async def __aiter__(self) -> AsyncGenerator[Event, None]:
        try:
            while True:
                yield await self.get()
        except Unsubscribed:
            pass

    async def get(self) -> Event:
        item = await self._queue.get()
        if item is None:
            raise Unsubscribed()
        return item


class MemoryBackend:
    def __init__(self):
        self._subscribed: typing.Set[str] = set()

    async def connect(self) -> None:
        self._published: asyncio.Queue[Event] = asyncio.Queue()

    async def disconnect(self) -> None:
        pass

    async def subscribe(self, channel: str) -> None:
        self._subscribed.add(channel)

    async def unsubscribe(self, channel: str) -> None:
        self._subscribed.remove(channel)

    async def publish(self, channel: str, message: typing.Any) -> None:
        event = Event(channel=channel, message=message)
        await self._published.put(event)

    async def next_published(self) -> Event:
        while True:
            event = await self._published.get()
            if event.channel in self._subscribed:
                return event


class Unsubscribed(Exception):
    pass


class Broadcast:
    def __init__(self):
        self._subscribers: Dict[str, Any] = {}
        self._backend = MemoryBackend()

    async def __aenter__(self) -> "Broadcast":
        await self.connect()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        await self._backend.connect()
        self._listener_task = asyncio.create_task(self._listener())

    async def disconnect(self) -> None:
        if self._listener_task.done():
            self._listener_task.result()
        else:
            self._listener_task.cancel()
        await self._backend.disconnect()

    async def _listener(self) -> None:
        while True:
            event = await self._backend.next_published()
            for queue in list(self._subscribers.get(event.channel, [])):
                await queue.put(event)

    async def publish(self, channel: str, message: Any) -> None:
        await self._backend.publish(channel, message)

    @asynccontextmanager
    async def subscribe(self, channel: str) -> AsyncIterator[Subscriber]:
        queue: asyncio.Queue[Event | None] = asyncio.Queue()

        try:
            if not self._subscribers.get(channel):
                await self._backend.subscribe(channel)
                self._subscribers[channel] = set([queue])
            else:
                self._subscribers[channel].add(queue)

            yield Subscriber(queue)

            self._subscribers[channel].remove(queue)
            if not self._subscribers.get(channel):
                del self._subscribers[channel]
                await self._backend.unsubscribe(channel)
        finally:
            await queue.put(None)



import os
from pathlib import Path
from unittest import TestCase

import pytest

from mentat.code_context import CodeContext, CodeContextSettings
from mentat.code_file import CodeMessageLevel
from mentat.config_manager import ConfigManager
from mentat.errors import UserError
from mentat.include_files import expand_paths
from mentat.llm_api import count_tokens


@pytest.mark.asyncio
async def test_path_gitignoring(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    gitignore_path = ".gitignore"
    testing_dir_path = "git_testing_dir"
    os.makedirs(testing_dir_path)

    # create 3 files, 2 ignored in gitignore, 1 not
    ignored_file_path_1 = Path(os.path.join(testing_dir_path, "ignored_file_1.txt"))
    ignored_file_path_2 = Path(os.path.join(testing_dir_path, "ignored_file_2.txt"))
    non_ignored_file_path = Path(os.path.join(testing_dir_path, "non_ignored_file.txt"))

    with open(gitignore_path, "w") as gitignore_file:
        gitignore_file.write("ignored_file_1.txt\nignored_file_2.txt")

    for file_path in [ignored_file_path_1, ignored_file_path_2, non_ignored_file_path]:
        with open(file_path, "w") as file:
            file.write("I am a file")

    # Run CodeFileManager on the git_testing_dir, and also explicitly pass in ignored_file_2.txt
    paths = [Path(testing_dir_path), Path(ignored_file_path_2)]
    mock_code_context.settings.paths = paths
    mock_code_context._set_include_files()

    expected_file_paths = [
        os.path.join(temp_testbed, ignored_file_path_2),
        os.path.join(temp_testbed, non_ignored_file_path),
    ]

    case = TestCase()
    file_paths = [
        str(file_path.resolve()) for file_path in mock_code_context.include_files
    ]
    case.assertListEqual(sorted(expected_file_paths), sorted(file_paths))


@pytest.mark.asyncio
async def test_config_glob_exclude(
    mocker, temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Makes sure glob exclude config works
    mock_glob_exclude = mocker.MagicMock()
    mocker.patch.object(ConfigManager, "file_exclude_glob_list", new=mock_glob_exclude)
    mock_glob_exclude.side_effect = [[os.path.join("glob_test", "**", "*.py")]]

    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.py")
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.ts")
    directly_added_glob_excluded_path = Path(
        os.path.join("glob_test", "bagel", "apple", "directly_added_glob_excluded.py")
    )
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    with open(
        directly_added_glob_excluded_path, "w"
    ) as directly_added_glob_excluded_file:
        directly_added_glob_excluded_file.write(
            "Config excludes me but I'm included if added directly"
        )

    mock_code_context.settings.paths = [Path("."), directly_added_glob_excluded_path]
    mock_code_context._set_include_files()

    file_paths = [
        str(file_path.resolve()) for file_path in mock_code_context.include_files
    ]
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_include_path) in file_paths
    assert os.path.join(temp_testbed, directly_added_glob_excluded_path) in file_paths


@pytest.mark.asyncio
async def test_glob_include(temp_testbed, mock_stream, mock_config, mock_code_context):
    # Make sure glob include works
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.py")
    glob_include_path2 = os.path.join("glob_test", "bagel", "apple", "include_me2.py")
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_path), exist_ok=True)
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    os.makedirs(os.path.dirname(glob_include_path2), exist_ok=True)
    with open(glob_include_path2, "w") as glob_include_file:
        glob_include_file.write("I am also included")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am not included")

    file_paths = expand_paths(["**/*.py"])
    mock_code_context.settings.paths = file_paths
    mock_code_context._set_include_files()

    file_paths = [
        str(file_path.resolve()) for file_path in mock_code_context.include_files
    ]
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_include_path) in file_paths
    assert os.path.join(temp_testbed, glob_include_path2) in file_paths


@pytest.mark.asyncio
async def test_cli_glob_exclude(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Make sure cli glob exclude works and overrides regular include
    glob_include_then_exclude_path = os.path.join(
        "glob_test", "bagel", "apple", "include_then_exclude_me.py"
    )
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_then_exclude_path), exist_ok=True)
    with open(glob_include_then_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am included then excluded")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")

    file_paths = expand_paths(["**/*.py"])
    exclude_paths = expand_paths(["**/*.py", "**/*.ts"])
    mock_code_context.settings.paths = file_paths
    mock_code_context.settings.exclude_paths = exclude_paths
    mock_code_context._set_include_files()

    file_paths = [file_path for file_path in mock_code_context.include_files]
    assert os.path.join(temp_testbed, glob_include_then_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths


@pytest.mark.asyncio
async def test_text_encoding_checking(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Makes sure we don't include non text encoded files, and we quit if user gives us one
    nontext_path = "iamnottext.py"
    with open(nontext_path, "wb") as f:
        # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
        f.write(bytearray([0x81]))

    mock_code_context.settings.paths = [Path("./")]
    mock_code_context._set_include_files()
    file_paths = [file_path for file_path in mock_code_context.include_files]
    assert os.path.join(temp_testbed, nontext_path) not in file_paths

    with pytest.raises(UserError) as e_info:
        nontext_path_requested = "iamalsonottext.py"
        with open(nontext_path_requested, "wb") as f:
            # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
            f.write(bytearray([0x81]))

        paths = [Path(nontext_path_requested)]
        mock_code_context.settings.paths = paths
        mock_code_context._set_include_files()
    assert e_info.type == UserError


@pytest.fixture
def features(mocker):
    features_meta = [
        ("somefile.txt", CodeMessageLevel.CODE, "some diff"),
        ("somefile.txt", CodeMessageLevel.CMAP_FULL, "some diff"),
        ("somefile.txt", CodeMessageLevel.CMAP_FULL, None),
        ("somefile.txt", CodeMessageLevel.CMAP, None),
        ("differentfile.txt", CodeMessageLevel.CODE, "some diff"),
    ]
    features = []
    for file, level, diff in features_meta:
        feature = mocker.MagicMock()
        feature.path = Path(file)
        feature.level = level
        feature.diff = diff
        features.append(feature)
    return features


@pytest.mark.asyncio
async def test_get_code_message_cache(
    mocker,
    temp_testbed,
    mock_config,
    mock_stream,
    mock_parser,
    mock_git_root,
    mock_code_file_manager,
):
    code_context_settings = CodeContextSettings(
        paths=["multifile_calculator"],
        exclude_paths=["multifile_calculator/calculator.py"],
        auto_tokens=10,
    )
    code_context = await CodeContext.create(code_context_settings)
    file = Path("multifile_calculator/operations.py")
    feature = mocker.MagicMock()
    feature.path = file
    code_context.features = [feature]

    # Return cached value if no changes to file or settings
    mock_get_code_message = mocker.patch(
        "mentat.code_context.CodeContext._get_code_message"
    )
    mock_get_code_message.return_value = "test1"
    value1 = await code_context.get_code_message("gpt-4", 1e6)
    mock_get_code_message.return_value = "test2"
    value2 = await code_context.get_code_message("gpt-4", 1e6)
    assert value1 == value2

    # Regenerate if settings change
    code_context.settings.auto_tokens = 11
    value3 = await code_context.get_code_message("gpt-4", 1e6)
    assert value1 != value3

    # Regenerate if feature files change
    mock_get_code_message.return_value = "test3"
    lines = file.read_text().splitlines()
    lines[0] = "something different"
    file.write_text("\n".join(lines))
    value4 = await code_context.get_code_message("gpt-4", 1e6)
    assert value3 != value4


@pytest.mark.asyncio
async def test_get_code_message_include(
    temp_testbed,
    mock_config,
    mock_parser,
    mock_git_root,
    mock_code_file_manager,
    mock_stream,
):
    code_context_settings = CodeContextSettings(
        paths=["multifile_calculator"],
        exclude_paths=["multifile_calculator/calculator.py"],
    )
    code_context = await CodeContext.create(code_context_settings)

    # If max tokens is less than include_files, return include_files without
    # raising and Exception (that's handled elsewhere)
    code_context.settings.auto_tokens = 0
    code_message = await code_context.get_code_message("gpt-4", 1e6)
    expected = [
        "Code Files:",
        "",
        "multifile_calculator/__init__.py",
        "1:",
        "",
        "multifile_calculator/operations.py",
        *[
            f"{i+1}:{line}"
            for i, line in enumerate(
                Path("multifile_calculator/operations.py").read_text().split("\n")
            )
        ],
    ]
    assert code_message.splitlines() == expected

    async def _count_auto_tokens_where(limit: int) -> int:
        code_context.settings.auto_tokens = limit
        code_message = await code_context.get_code_message("gpt-4", 1e6)
        return count_tokens(code_message, "gpt-4")

    # If max_tokens is None, include the full auto-context
    if not code_context.settings.no_code_map:
        assert await _count_auto_tokens_where(None) == 288  # Cmap w/ signatures
        assert await _count_auto_tokens_where(250) == 220  # Cmap
    assert await _count_auto_tokens_where(200) == 153  # fnames
    # Always return include_files, regardless of max
    assert await _count_auto_tokens_where(0) == 102  # Include_files only



import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Callable, DefaultDict, Dict, Set, cast

from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.completion.word_completer import WordCompleter
from prompt_toolkit.document import Document
from pygments.lexer import Lexer
from pygments.lexers import guess_lexer_for_filename
from pygments.token import Token
from pygments.util import ClassNotFound

from mentat.code_context import CODE_CONTEXT
from mentat.commands import Command


@dataclass
class SyntaxCompletion:
    words: Set[str]
    created_at: datetime = datetime.utcnow()


class MentatCompleter(Completer):
    def __init__(self):
        self.syntax_completions: Dict[Path, SyntaxCompletion] = dict()
        self.file_name_completions: DefaultDict[str, Set[Path]] = defaultdict(set)
        self.command_completer = WordCompleter(
            words=Command.get_command_completions(),
            ignore_case=True,
            sentence=True,
        )

        self._all_syntax_words: Set[str] = set()
        self._last_refresh_at: datetime | None = None

    def refresh_completions_for_file_path(self, file_path: Path):
        """Add/edit/delete completions for some filepath"""
        try:
            with open(file_path, "r") as f:
                file_content = f.read()
        except (FileNotFoundError, NotADirectoryError):
            logging.debug(f"Skipping {file_path}. Reason: file not found")
            return
        try:
            lexer = guess_lexer_for_filename(file_path, file_content)
            lexer = cast(Lexer, lexer)
        except ClassNotFound:
            logging.debug(f"Skipping {file_path}. Reason: lexer not found")
            return

        self.file_name_completions[file_path.name].add(file_path)

        tokens = list(lexer.get_tokens(file_content))
        filtered_tokens = set[str]()
        for token_type, token_value in tokens:
            if token_type not in Token.Name:
                continue
            if len(token_value) <= 1:
                continue
            filtered_tokens.add(token_value)
        self.syntax_completions[file_path] = SyntaxCompletion(words=filtered_tokens)

    async def refresh_completions(self):
        code_context = CODE_CONTEXT.get()

        file_paths = list(code_context.include_files.keys())

        # Remove syntax completions for files not in the context
        for file_path in set(self.syntax_completions.keys()):
            if file_path not in file_paths:
                del self.syntax_completions[file_path]
                file_name = file_path.name
                self.file_name_completions[file_name].remove(file_path)
                if len(self.file_name_completions[file_name]) == 0:
                    del self.file_name_completions[file_name]

        # Add/update syntax completions for files in the context
        for file_path in file_paths:
            if file_path not in self.syntax_completions:
                self.refresh_completions_for_file_path(file_path)
            else:
                modified_at = datetime.utcfromtimestamp(os.path.getmtime(file_path))
                if self.syntax_completions[file_path].created_at < modified_at:
                    self.refresh_completions_for_file_path(file_path)

        # Build de-duped syntax completions
        _all_syntax_words = set[str]()
        for syntax_completion in self.syntax_completions.values():
            _all_syntax_words.update(syntax_completion.words)
        self._all_syntax_words = _all_syntax_words

        self._last_refresh_at = datetime.utcnow()

    def get_completions(  # pyright: ignore
        self, document: Document, complete_event: CompleteEvent
    ):
        raise NotImplementedError

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ):
        if document.text_before_cursor == "":
            return
        if (
            self._last_refresh_at is None
            or (datetime.utcnow() - self._last_refresh_at).seconds > 5
        ):
            await self.refresh_completions()
        if (
            document.text_before_cursor[0] == "/"
            and not document.text_before_cursor[-1].isspace()
        ):
            command_completions = self.command_completer.get_completions(
                document, complete_event
            )
            for completion in command_completions:
                yield completion

        if document.text_before_cursor[-1] == " ":
            return
        document_words = document.text_before_cursor.split()
        if not document_words:
            return

        last_word = document_words[-1]
        get_completion_insert: Callable[[str], str] = lambda word: f"`{word}`"
        if last_word[0] == "`" and len(last_word) > 1:
            last_word = last_word.lstrip("`")
            get_completion_insert: Callable[[str], str] = lambda word: f"{word}`"

        completions = self._all_syntax_words.union(set(self.file_name_completions))

        for completion in completions:
            if completion.lower().startswith(last_word.lower()):
                file_names = self.file_name_completions.get(completion)
                if file_names:
                    for file_name in file_names:
                        yield Completion(
                            get_completion_insert(str(file_name)),
                            start_position=-len(last_word),
                            display=str(file_name),
                        )
                else:
                    yield Completion(
                        get_completion_insert(completion),
                        start_position=-len(last_word),
                        display=completion,
                    )



from __future__ import annotations

import asyncio
import os
from contextvars import ContextVar
from pathlib import Path
from textwrap import dedent
from typing import Optional

import attr

from .code_file import CodeFile, CodeMessageLevel
from .code_file_manager import CODE_FILE_MANAGER
from .code_map import check_ctags_disabled
from .diff_context import DiffContext
from .errors import UserError
from .git_handler import GIT_ROOT, get_non_gitignored_files, get_paths_with_git_diffs
from .include_files import (
    build_path_tree,
    get_include_files,
    is_file_text_encoded,
    print_path_tree,
)
from .llm_api import count_tokens
from .session_stream import SESSION_STREAM
from .utils import sha256


async def _count_tokens_in_features(features: list[CodeFile], model: str) -> int:
    sem = asyncio.Semaphore(10)

    async def _count_tokens(feature: CodeFile) -> int:
        async with sem:
            return await feature.count_tokens(model)

    tasks = [_count_tokens(f) for f in features]
    results = await asyncio.gather(*tasks)
    return sum(results)


@attr.define
class CodeContextSettings:
    paths: list[Path] = []
    exclude_paths: list[Path] = []
    diff: Optional[str] = None
    pr_diff: Optional[str] = None
    no_code_map: bool = False
    auto_tokens: Optional[int] = None


CODE_CONTEXT: ContextVar[CodeContext] = ContextVar("mentat:code_context")


class CodeContext:
    settings: CodeContextSettings
    include_files: dict[Path, CodeFile]
    diff_context: DiffContext
    code_map: bool = True
    features: list[CodeFile] = []

    def __init__(self, settings: CodeContextSettings):
        self.settings = settings

    @classmethod
    async def create(cls, settings: CodeContextSettings):
        self = cls(settings)

        await self._set_diff_context(replace_paths=True)
        self._set_include_files()
        await self._set_code_map()

        return self

    async def _set_diff_context(self, replace_paths: bool = False):
        try:
            self.diff_context = DiffContext.create(
                self.settings.diff, self.settings.pr_diff
            )
            if replace_paths and not self.settings.paths:
                self.settings.paths = self.diff_context.files
        except UserError as e:
            await SESSION_STREAM.get().send(str(e), color="light_yellow")
            exit()

    def _set_include_files(self):
        self.include_files = get_include_files(
            self.settings.paths, self.settings.exclude_paths
        )

    async def _set_code_map(self):
        if self.settings.no_code_map:
            self.code_map = False
        else:
            disabled_reason = await check_ctags_disabled()
            if disabled_reason:
                ctags_disabled_message = f"""
                    There was an error with your universal ctags installation, disabling CodeMap.
                    Reason: {disabled_reason}
                """
                ctags_disabled_message = dedent(ctags_disabled_message)
                await SESSION_STREAM.get().send(ctags_disabled_message, color="yellow")
                self.settings.no_code_map = True
                self.code_map = False
            else:
                self.code_map = True

    async def display_context(self):
        """Display the baseline context: included files and auto-context settings"""
        stream = SESSION_STREAM.get()
        git_root = GIT_ROOT.get()

        await stream.send("\nCode Context:", color="blue")
        prefix = "  "
        await stream.send(f"{prefix}Directory: {git_root}")
        if self.diff_context.name:
            await stream.send(f"{prefix}Diff:", end=" ")
            await stream.send(self.diff_context.get_display_context(), color="green")
        if self.include_files:
            await stream.send(f"{prefix}Included files:")
            await stream.send(f"{prefix + prefix}{git_root.name}")
            await print_path_tree(
                build_path_tree(list(self.include_files.values()), git_root),
                get_paths_with_git_diffs(),
                git_root,
                prefix + prefix,
            )
        else:
            await stream.send(f"{prefix}Included files: None", color="yellow")
        await stream.send(
            f"{prefix}CodeMaps: {'Enabled' if self.code_map else 'Disabled'}"
        )
        auto = self.settings.auto_tokens
        await stream.send(
            f"{prefix}Auto-tokens: {'Model max (default)' if auto is None else auto}"
        )

    async def display_features(self):
        """Display a summary of all active features"""
        auto_features = {level: 0 for level in CodeMessageLevel}
        for f in self.features:
            if f.path not in self.include_files:
                auto_features[f.level] += 1
        if any(auto_features.values()):
            stream = SESSION_STREAM.get()
            await stream.send("Auto-Selected Features:", color="blue")
            for level, count in auto_features.items():
                if count:
                    await stream.send(f"  {count} {level.description}")

    _code_message: str | None = None
    _code_message_checksum: str | None = None

    def _get_code_message_checksum(self, max_tokens: Optional[int] = None) -> str:
        if not self.features:
            features_checksum = ""
        else:
            git_root = GIT_ROOT.get()
            code_file_manager = CODE_FILE_MANAGER.get()
            feature_files = {Path(git_root / f.path) for f in self.features}
            feature_file_checksums = [
                code_file_manager.get_file_checksum(f) for f in feature_files
            ]
            features_checksum = sha256("".join(feature_file_checksums))
        settings = attr.asdict(self.settings)
        settings["max_tokens"] = max_tokens
        settings_checksum = sha256(str(settings))
        return features_checksum + settings_checksum

    async def get_code_message(
        self,
        model: str,
        max_tokens: int,
    ) -> str:
        code_message_checksum = self._get_code_message_checksum(max_tokens)
        if (
            self._code_message is None
            or code_message_checksum != self._code_message_checksum
        ):
            self._code_message = await self._get_code_message(model, max_tokens)
            self._code_message_checksum = self._get_code_message_checksum(max_tokens)
        return self._code_message

    async def _get_code_message(
        self,
        model: str,
        max_tokens: int,
    ) -> str:
        code_message = list[str]()

        await self._set_diff_context()
        self._set_include_files()
        await self._set_code_map()
        if self.diff_context.files:
            code_message += [
                "Diff References:",
                f' "-" = {self.diff_context.name}',
                ' "+" = Active Changes',
                "",
            ]
        code_message += ["Code Files:\n"]

        features = self._get_include_features()
        include_feature_tokens = await _count_tokens_in_features(
            features, model
        ) - count_tokens("\n".join(code_message), model)
        _max_auto = max(0, max_tokens - include_feature_tokens)
        _max_user = self.settings.auto_tokens
        if _max_auto == 0 or _max_user == 0:
            self.features = features
        else:
            auto_tokens = _max_auto if _max_user is None else min(_max_auto, _max_user)
            self.features = await self._get_auto_features(model, features, auto_tokens)

        for f in self.features:
            code_message += await f.get_code_message()
        return "\n".join(code_message)

    def _get_include_features(self) -> list[CodeFile]:
        git_root = GIT_ROOT.get()
        include_features = list[CodeFile]()
        for path, feature in self.include_files.items():
            if feature.level == CodeMessageLevel.INTERVAL:
                interval_str = ",".join(f"{i.start}-{i.end}" for i in feature.intervals)
                path = f"{path}:{interval_str}"
            diff_target = (
                self.diff_context.target if path in self.diff_context.files else None
            )
            feature = CodeFile(path, feature.level, diff=diff_target)
            include_features.append(feature)

        def _feature_relative_path(f: CodeFile) -> str:
            return os.path.relpath(f.path, git_root)

        return sorted(include_features, key=_feature_relative_path)

    async def _get_auto_features(
        self,
        model: str,
        include_features: list[CodeFile],
        max_tokens: int,
    ) -> list[CodeFile]:
        git_root = GIT_ROOT.get()

        # Find the first (longest) level that fits
        include_features_tokens = await _count_tokens_in_features(
            include_features, model
        )
        max_auto_tokens = max_tokens - include_features_tokens
        all_features = include_features.copy()
        levels = [CodeMessageLevel.FILE_NAME]
        if not self.settings.no_code_map:
            levels = [CodeMessageLevel.CMAP_FULL, CodeMessageLevel.CMAP] + levels
        for level in levels:
            _features = list[CodeFile]()
            for path in get_non_gitignored_files(git_root):
                if (
                    path in self.include_files
                    or path.is_dir()
                    or not is_file_text_encoded(path)
                ):
                    continue
                diff_target = (
                    self.diff_context.target
                    if path in self.diff_context.files
                    else None
                )
                feature = CodeFile(path, level=level, diff=diff_target)
                _features.append(feature)
            level_length = await _count_tokens_in_features(_features, model)
            if level_length < max_auto_tokens:
                all_features += _features
                break

        def _feature_relative_path(f: CodeFile) -> str:
            return os.path.relpath(f.path, git_root)

        return sorted(all_features, key=_feature_relative_path)

    async def include_file(self, code_file: CodeFile):
        stream = SESSION_STREAM.get()
        if not os.path.exists(code_file.path):
            await stream.send(f"File does not exist: {code_file.path}\n", color="red")
            return
        if code_file.path in self.settings.paths:
            await stream.send(
                f"File already in context: {code_file.path}\n", color="yellow"
            )
            return
        if code_file.path in self.settings.exclude_paths:
            self.settings.exclude_paths.remove(code_file.path)
        self.settings.paths.append(code_file.path)
        self._set_include_files()
        await stream.send(
            f"File included in context: {code_file.path}\n", color="green"
        )

    async def exclude_file(self, code_file: CodeFile):
        stream = SESSION_STREAM.get()
        if not os.path.exists(code_file.path):
            await stream.send(f"File does not exist: {code_file.path}\n", color="red")
            return
        if code_file.path not in self.settings.paths:
            await stream.send(
                f"File not in context: {code_file.path}\n", color="yellow"
            )
            return
        if code_file.path in self.settings.exclude_paths:
            self.settings.paths.remove(code_file.path)
        self.settings.exclude_paths.append(code_file.path)
        self._set_include_files()
        await stream.send(
            f"File removed from context: {code_file.path}\n", color="green"
        )



import asyncio
import json
import os
import webbrowser
from functools import partial
from multiprocessing import Pool

import pytest
import tqdm
from git import Repo
from openai import InvalidRequestError

from mentat.llm_api import COST_TRACKER, call_llm_api, setup_api_key
from mentat.python_client.client import PythonClient

from .exercise_runners.exercise_runner_factory import ExerciseRunnerFactory

pytestmark = pytest.mark.benchmark


@pytest.fixture
def clone_exercism_repo(refresh_repo, language):
    exercism_url = f"https://github.com/exercism/{language}.git"

    local_dir = f"{os.path.dirname(__file__)}/../../../exercism-{language}"
    if os.path.exists(local_dir):
        if refresh_repo:
            repo = Repo(local_dir)
            repo.git.reset("--hard")
            repo.git.clean("-fd")
            repo.remotes.origin.pull()
    else:
        repo = Repo.clone_from(exercism_url, local_dir)

    os.chdir(local_dir)


@pytest.fixture
def exercises(request):
    exercises = request.config.getoption("--exercises")
    if len(exercises) == 1:
        return exercises[0]
    return exercises


@pytest.fixture
def max_exercises(request):
    return int(request.config.getoption("--max_exercises"))


@pytest.fixture
def max_iterations(request):
    return int(request.config.getoption("--max_iterations"))


@pytest.fixture
def refresh_repo(request):
    return request.config.getoption("--refresh_repo")


@pytest.fixture
def max_workers(request):
    return int(request.config.getoption("--max_workers"))


@pytest.fixture
def language(request):
    return request.config.getoption("--language")


prompt = (
    "You are a professional code reviewer who helps other coders improve their skills."
    " You recently assigned a coder a small coding test to assess their level, with a"
    " pre-written stub template and an automated test suite to determine if they"
    " succeeded. They failed the test; using the instructions for the test, the code"
    " they wrote for the test, and the output of the test suite, your job is to"
    " determine why they failed. Give a terse but informative couple of sentences on"
    " why they failed, before giving the reason it failed on the final line in the"
    " format:\n"
    + "reason: <reason_failed>\n"
    + "Your response will be parsed programmatically, so you MUST follow the format for"
    " the final line! The possible responses for the final line and what they mean"
    " are as follows:\n"
    + "blank (the coder didn't change the file at all from the stub you provided"
    " them)\n"
    + "wording (everything was correct, but the coder messed up the wording or spacing"
    " which caused it to be rejected)\n"
    + "duplication (the coder had a random duplicated line that caused the code to not"
    " be compiled/interpreted)\n"
    + "syntax (the coder messed up their syntax, meaning their code couldn't be"
    " compiled/interpreted)\n"
    + "logic (the coder messed up the logic)\n"
    + "other (some other reason caused it to fail)\n"
)
model = "gpt-4-0314"


async def failure_analysis(exercise_runner, language):
    setup_api_key()

    instructions = exercise_runner.read_instructions()
    code = exercise_runner.read_code(language)
    test_results = exercise_runner.read_test_results()

    final_message = (
        f"All instructions:\n{instructions}\nCode to review:\n{code}\nTest"
        f" results:\n{test_results}"
    )
    messages = [
        {"role": "system", "content": prompt},
        {"role": "user", "content": final_message},
    ]
    response = ""
    try:
        async for chunk in await call_llm_api(messages, model):
            content = chunk["choices"][0]["delta"].get("content", "")
            response += content
    except InvalidRequestError:
        response = "Unable to analyze test case\nreason: too many tokens to analyze"

    try:
        reason = response.strip().split("\n")[-1][8:]
    except IndexError:
        reason = "response error"
    return response, reason


async def run_exercise(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    old_result = exercise_runner.get_result_from_txt()
    if old_result:
        return old_result
    client = PythonClient(
        paths=exercise_runner.include_files(),
        exclude_paths=exercise_runner.exclude_files(),
        no_code_map=True,
    )

    prompt_1 = (
        f"Use the instructions in {exercise_runner.docs()} to modify"
        + f" {exercise_runner.file}. Keep and implement the existing"
        + " function or class stubs, they will be called from unit tests. Only use"
        + f" standard {language} libraries, don't suggest installing any packages."
    )
    prompt_2 = (
        "\nSee the testing errors above. The tests are correct. Fix the code"
        + f" in {exercise_runner.file} to resolve the errors."
    )

    iterations = 0
    while iterations < max_iterations:
        if exercise_runner.passed():
            break
        message = (
            prompt_1
            if iterations == 0
            else exercise_runner.get_error_message() + prompt_2
        )
        await client.call_mentat_auto_accept(message)

        exercise_runner.run_test()
        iterations += 1

    await client.stop()
    passed = exercise_runner.passed()
    result = {
        "iterations": iterations,
        "passed": passed,
        "test": exercise_runner.name,
    }
    if not result["passed"]:
        response, reason = await failure_analysis(exercise_runner, language)
        result["response"] = response
        result["reason"] = reason
    result["tokens"] = COST_TRACKER.get().total_tokens
    return result


def run_exercise_sync(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    try:
        result = asyncio.run(run_exercise(problem_dir, language, max_iterations))
    except Exception as e:
        print(f"\nError running {problem_dir}")
        print(str(e), flush=True)
        result = {
            "iterations": 0,
            "passed": False,
            "test": problem_dir,
            "response": str(e),
            "reason": "error",
            "tokens": 0,
        }
    result["instructions"] = exercise_runner.read_instructions()
    result["code"] = exercise_runner.read_code(language)
    result["test-output"] = exercise_runner.read_test_results()
    return result


def summarize_results(results):
    passed_in_n = {}
    failed = 0
    for result in results:
        if result["passed"]:
            iteration = result["iterations"]
            if iteration:
                passed_in_n[iteration] = passed_in_n.get(iteration, 0) + 1
        else:
            failed += 1
    return "Passed: " + str(passed_in_n)[1:-1] + "| Failed: " + str(failed)


def test_practice_directory_performance(
    clone_exercism_repo,
    exercises,
    max_exercises,
    max_iterations,
    max_workers,
    language,
):
    all_exercises = os.listdir("exercises/practice")
    if len(exercises) > 0:
        exercises = set(exercises) & set(all_exercises)
    else:
        exercises = all_exercises[:max_exercises]
    num_exercises = len(exercises)

    # TODO: aiomultiprocessing would be faster with fewer workers; setup a Manager in a parent process
    # that controls the children processes so that we don't run into rate limits
    with Pool(processes=max_workers) as pool:
        pbar = tqdm.tqdm(total=num_exercises)

        result_map = pool.imap(
            partial(
                run_exercise_sync, language=language, max_iterations=max_iterations
            ),
            exercises,
        )
        results = []
        for result in result_map:
            pbar.update()
            results.append(result)
            with open("results.txt", "a") as f:
                json.dump(result, f)
                f.write("\n")
            pbar.set_description(
                summarize_results(results) + "| Last Ran: " + result["test"]
            )
        results.sort(key=lambda result: result["test"])

        # Update the html file
        results_json = list(map(json.dumps, results))
        results_str = "[" + ",".join(results_json) + "]"
        with open(f"{os.path.dirname(__file__)}/exercism_benchmark.html", "r") as f:
            html = f.read()
        html = html.replace("{{ results }}", results_str)
        with open("results.html", "w") as f:
            f.write(html)
        webbrowser.open("file://" + os.path.realpath("results.html"))



import subprocess
from pathlib import Path

import pytest

from mentat.commands import (
    Command,
    ExcludeCommand,
    HelpCommand,
    IncludeCommand,
    InvalidCommand,
)


def test_invalid_command():
    assert isinstance(Command.create_command("non-existent"), InvalidCommand)


@pytest.mark.asyncio
async def test_help_command(mock_stream):
    command = Command.create_command("help")
    await command.apply()
    assert isinstance(command, HelpCommand)


@pytest.mark.asyncio
async def test_commit_command(mock_stream, temp_testbed):
    file_name = "test_file.py"
    with open(file_name, "w") as f:
        f.write("# Commit me!")

    command = Command.create_command("commit")
    await command.apply("commit", "test_file committed")
    assert subprocess.check_output(["git", "diff", "--name-only"], text=True) == ""


@pytest.mark.asyncio
async def test_include_command(mock_stream, mock_config, mock_code_context):
    command = Command.create_command("include")
    assert isinstance(command, IncludeCommand)
    await command.apply("__init__.py")
    assert Path("__init__.py").resolve() in mock_code_context.include_files


@pytest.mark.asyncio
async def test_exclude_command(mock_stream, mock_config, mock_code_context):
    command = Command.create_command("exclude")
    assert isinstance(command, ExcludeCommand)
    await command.apply("__init__.py")
    assert Path("__init__.py").resolve() not in mock_code_context.include_files



from __future__ import annotations

import json
import logging
from contextvars import ContextVar
from importlib import resources
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Optional, cast

from mentat.git_handler import GIT_ROOT
from mentat.session_stream import SESSION_STREAM

mentat_dir_path = Path.home() / ".mentat"

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
default_config_file_name = "default_config.json"
config_file_name = ".mentat_config.json"
user_config_path = mentat_dir_path / config_file_name

CONFIG_MANAGER: ContextVar[ConfigManager] = ContextVar("mentat:config_manager")


class ConfigManager:
    def __init__(
        self,
        user_config: Dict[str, str],
        project_config: Dict[str, str],
    ):
        self.user_config = user_config
        self.project_config = project_config

        default_config_path = resources.files(package_name).joinpath(
            default_config_file_name
        )
        with default_config_path.open("r") as config_file:
            self.default_config = json.load(config_file)

    @classmethod
    async def create(cls):
        stream = SESSION_STREAM.get()

        if user_config_path.exists():
            with open(user_config_path) as config_file:
                try:
                    user_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("User config file contains invalid json")
                    await stream.send(
                        "Warning: User .mentat_config.json contains invalid"
                        " json; ignoring user configuration file",
                        color="light_yellow",
                    )
                    user_config = dict[str, str]()
        else:
            user_config = dict[str, str]()

        project_config_path = GIT_ROOT.get() / config_file_name
        if project_config_path.exists():
            with open(project_config_path) as config_file:
                try:
                    project_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("Project config file contains invalid json")
                    await stream.send(
                        "Warning: Git project .mentat_config.json contains invalid"
                        " json; ignoring project configuration file",
                        color="light_yellow",
                    )
                    project_config = dict[str, str]()
        else:
            project_config = dict[str, str]()

        self = cls(user_config, project_config)

        return self

    def input_style(self) -> list[tuple[str, str]]:
        return cast(list[tuple[str, str]], self._get_key("input-style"))

    def model(self) -> str:
        return cast(str, self._get_key("model"))

    def maximum_context(self) -> Optional[int]:
        maximum_context = self._get_key("maximum-context")
        if maximum_context:
            return int(maximum_context)
        return None

    def file_exclude_glob_list(self) -> list[str]:
        return cast(list[str], self._get_key("file-exclude-glob-list"))

    def parser(self) -> str:
        return cast(str, self._get_key("format"))

    def _get_key(self, key: str) -> Any:
        if key in self.project_config:
            return self.project_config[key]
        elif key in self.user_config:
            return self.user_config[key]
        elif key in self.default_config:
            return self.default_config[key]
        else:
            raise ValueError(f"No value for config key {key} found")



import glob
import logging
import os
from pathlib import Path
from typing import Any, Dict

from termcolor import cprint

from mentat.code_file import CodeFile, parse_intervals
from mentat.config_manager import CONFIG_MANAGER
from mentat.errors import UserError
from mentat.git_handler import GIT_ROOT, get_non_gitignored_files
from mentat.session_stream import SESSION_STREAM


def expand_paths(paths: list[str]) -> list[Path]:
    """Expand user-input paths/globs into a list of absolute paths.

    Should be done as soon as possible because some shells such as zsh automatically
    expand globs and we want to avoid differences in functionality between shells
    """
    globbed_paths = set[str]()
    invalid_paths = list[str]()
    for path in paths:
        new_paths = glob.glob(pathname=path, recursive=True)
        if new_paths:
            globbed_paths.update(new_paths)
        else:
            split = path.rsplit(":", 1)
            p = split[0]
            if len(split) > 1:
                # Parse additional syntax, e.g. "path/to/file.py:1-5,7,12-40"
                intervals = parse_intervals(split[1])
            else:
                intervals = None
            if Path(p).exists() and intervals:
                globbed_paths.add(path)
            else:
                invalid_paths.append(path)
    if invalid_paths:
        cprint(
            "The following paths do not exist:",
            "light_yellow",
        )
        print("\n".join(invalid_paths))
        exit()
    return [Path(path) for path in globbed_paths]


def is_file_text_encoded(file_path: Path):
    """Checks if a file is text encoded."""
    try:
        # The ultimate filetype test
        with open(file_path, "r") as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False


def abs_files_from_list(paths: list[Path], check_for_text: bool = True):
    """Returns a set of CodeFiles from a list of paths."""
    files_direct = set[CodeFile]()
    file_paths_from_dirs = set[Path]()
    for path in paths:
        file = CodeFile(os.path.realpath(path))
        path = Path(file.path)
        if path.is_file():
            if check_for_text and not is_file_text_encoded(path):
                logging.info(f"File path {path} is not text encoded.")
                raise UserError(f"File path {path} is not text encoded.")
            files_direct.add(file)
        elif path.is_dir():
            nonignored_files = set(
                map(
                    lambda f: Path(os.path.realpath(path / f)),
                    get_non_gitignored_files(path),
                )
            )

            file_paths_from_dirs.update(
                filter(
                    lambda f: (not check_for_text) or is_file_text_encoded(f),
                    nonignored_files,
                )
            )

    files_from_dirs = [
        CodeFile(os.path.realpath(path)) for path in file_paths_from_dirs
    ]
    return files_direct, files_from_dirs


def get_include_files(
    paths: list[Path], exclude_paths: list[Path]
) -> Dict[Path, CodeFile]:
    """Returns a complete list of text files in a given set of include/exclude Paths."""
    git_root = GIT_ROOT.get()
    config = CONFIG_MANAGER.get()
    excluded_files_direct, excluded_files_from_dirs = abs_files_from_list(
        exclude_paths, check_for_text=False
    )
    excluded_files, excluded_files_from_dir = set(
        map(lambda f: f.path, excluded_files_direct)
    ), set(map(lambda f: f.path, excluded_files_from_dirs))

    glob_excluded_files = set(
        os.path.join(git_root, file)
        for glob_path in config.file_exclude_glob_list()
        # If the user puts a / at the beginning, it will try to look in root directory
        for file in glob.glob(
            pathname=glob_path,
            root_dir=git_root,
            recursive=True,
        )
    )
    files_direct, files_from_dirs = abs_files_from_list(paths, check_for_text=True)

    # config glob excluded files only apply to files added from directories
    files_from_dirs = [
        file
        for file in files_from_dirs
        if str(file.path.resolve()) not in glob_excluded_files
    ]

    files_direct.update(files_from_dirs)

    files = dict[Path, CodeFile]()
    for file in files_direct:
        if file.path not in excluded_files | excluded_files_from_dir:
            files[Path(os.path.realpath(file.path))] = file

    return files


def build_path_tree(files: list[CodeFile], git_root: Path):
    """Builds a tree of paths from a list of CodeFiles."""
    tree = dict[str, Any]()
    for file in files:
        path = os.path.relpath(file.path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


async def print_path_tree(
    tree: dict[str, Any], changed_files: set[Path], cur_path: Path, prefix: str = ""
):
    """Prints a tree of paths, with changed files highlighted."""
    stream = SESSION_STREAM.get()
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            await stream.send(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            await stream.send(f"{prefix}└── ", end="")

        cur = cur_path / key
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif star:
            color = "green"
        else:
            color = None
        await stream.send(f"{star}{key}", color=color)
        if tree[key]:
            await print_path_tree(tree[key], changed_files, cur, new_prefix)



from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Event
from contextlib import asynccontextmanager
from contextvars import ContextVar
from pathlib import Path
from typing import Any, AsyncGenerator

from termcolor import colored

from mentat.code_file_manager import CODE_FILE_MANAGER, CodeFileManager
from mentat.errors import ModelError
from mentat.git_handler import GIT_ROOT
from mentat.llm_api import chunk_to_lines
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    FileActionType,
    change_delimiter,
    get_file_name,
    get_later_lines,
    get_previous_lines,
    get_removed_lines,
)
from mentat.parsers.file_edit import FileEdit
from mentat.session_stream import SESSION_STREAM
from mentat.streaming_printer import StreamingPrinter

PARSER: ContextVar[Parser] = ContextVar("mentat:parser")


class Parser(ABC):
    def __init__(self):
        self.shutdown = Event()
        self._interrupt_task = None

    async def listen_for_interrupt(self):
        stream = SESSION_STREAM.get()
        async with stream.interrupt_lock:
            await stream.recv("interrupt")
            logging.info("User interrupted response.")
            self.shutdown.set()

    @asynccontextmanager
    async def interrupt_catcher(self):
        self._interrupt_task = asyncio.create_task(self.listen_for_interrupt())
        yield
        if self._interrupt_task is not None:  # type: ignore
            self._interrupt_task.cancel()
            try:
                await self._interrupt_task
            except asyncio.CancelledError:
                pass
        self._interrupt_task = None
        self.shutdown.clear()

    @abstractmethod
    def get_system_prompt(self) -> str:
        pass

    async def stream_and_parse_llm_response(
        self,
        response: AsyncGenerator[Any, None],
    ) -> tuple[str, list[FileEdit]]:
        """
        This general parsing structure relies on the assumption that all formats require three types of lines:
        1. 'conversation' lines, which are streamed as they come,
        2. 'special' lines, that are never shown to the user and contain information such as the file_name
        3. 'code' lines, which are the actual code written and are shown to the user in a special format
        To make a parser that differs from these assumptions, make this functionality a subclass of Parser
        """

        stream = SESSION_STREAM.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        git_root = GIT_ROOT.get()

        printer = StreamingPrinter()
        printer_task = asyncio.create_task(printer.print_lines())
        message = ""
        file_edits = dict[Path, FileEdit]()

        cur_line = ""
        prev_block = ""
        cur_block = ""
        display_information: DisplayInformation | None = None
        file_edit: FileEdit | None = None
        line_printed = False
        in_special_lines = False
        in_code_lines = False
        conversation = True
        printed_delimiter = False
        rename_map = dict[Path, Path]()
        async for chunk in response:
            if self.shutdown.is_set():
                printer.shutdown_printer()
                await printer_task
                await stream.send(
                    colored("")  # Reset ANSI codes
                    + "\n\nInterrupted by user. Using the response up to this point."
                )
                break

            for content in chunk_to_lines(chunk):
                if not content:
                    continue
                message += content
                cur_line += content

                # Print if not in special lines and line is confirmed not special
                if not in_special_lines:
                    if not line_printed:
                        if not self._could_be_special(cur_line):
                            line_printed = True
                            to_print = (
                                cur_line
                                if not in_code_lines or display_information is None
                                else self._code_line_beginning(
                                    display_information, cur_block
                                )
                                + self._code_line_content(
                                    display_information, cur_line, cur_line, cur_block
                                )
                            )
                            printer.add_string(to_print, end="")
                    else:
                        to_print = (
                            content
                            if not in_code_lines or display_information is None
                            else self._code_line_content(
                                display_information, content, cur_line, cur_block
                            )
                        )
                        printer.add_string(to_print, end="")

                # If we print non code lines, we want to reprint the file name of the next change,
                # even if it's the same file as the last change
                if not in_code_lines and not in_special_lines and line_printed:
                    conversation = True

                # New line handling
                if "\n" in cur_line:
                    # Now that full line is in, give _could_be_special full line (including newline)
                    # and see if it should be printed or not
                    if (
                        not in_special_lines
                        and not line_printed
                        and not self._could_be_special(cur_line)
                    ):
                        to_print = (
                            cur_line
                            if not in_code_lines or display_information is None
                            else self._code_line_beginning(
                                display_information, cur_block
                            )
                            + self._code_line_content(
                                display_information, cur_line, cur_line, cur_block
                            )
                        )
                        printer.add_string(to_print, end="")
                        line_printed = True

                    if self._starts_special(cur_line.strip()):
                        in_special_lines = True

                    if in_special_lines or in_code_lines:
                        cur_block += cur_line

                    if in_special_lines and self._ends_special(cur_line.strip()):
                        previous_file = (
                            None if file_edit is None else file_edit.file_path
                        )

                        try:
                            (
                                display_information,
                                file_edit,
                                in_code_lines,
                            ) = self._special_block(
                                code_file_manager, git_root, rename_map, cur_block
                            )
                        except ModelError as e:
                            printer.add_string(str(e), color="red")
                            printer.add_string("Using existing changes.")
                            printer.wrap_it_up()
                            await printer_task
                            logging.debug("LLM Response:")
                            logging.debug(message)
                            return (
                                message,
                                [file_edit for file_edit in file_edits.values()],
                            )

                        in_special_lines = False
                        prev_block = cur_block
                        cur_block = ""

                        # Rename map handling
                        if display_information.new_name is not None:
                            rename_map[display_information.new_name] = (
                                display_information.file_name
                            )
                        if display_information.file_name in rename_map:
                            file_edit.file_path = (
                                git_root / rename_map[display_information.file_name]
                            )

                        # New file_edit creation and merging
                        if file_edit.file_path not in file_edits:
                            file_edits[file_edit.file_path] = file_edit
                        else:
                            cur_file_edit = file_edits[file_edit.file_path]
                            cur_file_edit.is_creation = (
                                cur_file_edit.is_creation or file_edit.is_creation
                            )
                            cur_file_edit.is_deletion = (
                                cur_file_edit.is_deletion or file_edit.is_deletion
                            )
                            if file_edit.rename_file_path is not None:
                                cur_file_edit.rename_file_path = (
                                    file_edit.rename_file_path
                                )
                            file_edit = cur_file_edit

                        # Print file header
                        if (
                            conversation
                            or display_information.file_action_type
                            == FileActionType.RenameFile
                            or (file_edit.file_path != previous_file)
                        ):
                            conversation = False
                            printer.add_string(get_file_name(display_information))
                            if in_code_lines or display_information.removed_block:
                                printed_delimiter = True
                                printer.add_string(change_delimiter)
                            else:
                                printed_delimiter = False
                        elif not printed_delimiter:
                            # We have to have this so that putting a change like an insert after a rename
                            # still has a change delimiter
                            printer.add_string(change_delimiter)
                            printed_delimiter = True

                        # Print previous lines, removed block, and possibly later lines
                        if in_code_lines or display_information.removed_block:
                            printer.add_string(get_previous_lines(display_information))
                            printer.add_string(get_removed_lines(display_information))
                            if not in_code_lines:
                                printer.add_string(get_later_lines(display_information))
                                printer.add_string(change_delimiter)
                    elif in_code_lines and self._ends_code(cur_line.strip()):
                        # Adding code lines to previous file_edit and printing later lines
                        if display_information is not None and file_edit is not None:
                            to_display = self._add_code_block(
                                code_file_manager,
                                rename_map,
                                prev_block,
                                cur_block,
                                display_information,
                                file_edit,
                            )
                            printer.add_string(get_later_lines(display_information))
                            printer.add_string(change_delimiter)
                            printer.add_string(to_display)
                        else:
                            printer.add_string(change_delimiter)

                        in_code_lines = False
                        prev_block = cur_block
                        cur_block = ""
                    line_printed = False
                    cur_line = ""
        else:
            # If the model doesn't close out the code lines, we might as well do it for it
            if (
                in_code_lines
                and display_information is not None
                and file_edit is not None
            ):
                to_display = self._add_code_block(
                    code_file_manager,
                    rename_map,
                    prev_block,
                    cur_block,
                    display_information,
                    file_edit,
                )
                printer.add_string(get_later_lines(display_information))
                printer.add_string(change_delimiter)
                printer.add_string(to_display)

            # Only finish printing if we don't quit from ctrl-c
            printer.wrap_it_up()
            await printer_task

        logging.debug("LLM Response:")
        logging.debug(message)
        return (message, [file_edit for file_edit in file_edits.values()])

    # Ideally this would be called in this class instead of subclasses
    def _get_file_lines(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        rel_path: Path,
    ) -> list[str]:
        path = rename_map.get(
            rel_path,
            rel_path,
        )
        return code_file_manager.file_lines.get(path, [])

    # These methods aren't abstract, since most parsers will use this implementation, but can be overriden easily
    def provide_line_numbers(self) -> bool:
        return True

    def _code_line_beginning(
        self, display_information: DisplayInformation, cur_block: str
    ) -> str:
        """
        The beginning of a code line; normally this means printing the + prefix
        """
        return colored(
            "+" + " " * (display_information.line_number_buffer - 1), color="green"
        )

    def _code_line_content(
        self,
        display_information: DisplayInformation,
        content: str,
        cur_line: str,
        cur_block: str,
    ) -> str:
        """
        Part of a code line; normally this means printing in green
        """
        return colored(content, color="green")

    @abstractmethod
    def _could_be_special(self, cur_line: str) -> bool:
        """
        Returns if this current line could be a special line and therefore shouldn't be printed yet.
        Once line is completed, will include a newline character.
        """
        pass

    @abstractmethod
    def _starts_special(self, line: str) -> bool:
        """
        Determines if this line begins a special block
        """
        pass

    @abstractmethod
    def _ends_special(self, line: str) -> bool:
        """
        Determines if this line ends a special block
        """
        pass

    @abstractmethod
    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        git_root: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        """
        After finishing special block, return DisplayInformation to print, FileEdit to add/merge to list,
        and if a code block follows this special block.
        """
        pass

    @abstractmethod
    def _ends_code(self, line: str) -> bool:
        """
        Determines if this line ends a code block
        """
        pass

    @abstractmethod
    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ) -> str:
        """
        Using the special block, code block and display_information, edits the FileEdit to add the new code block.
        Can return a message to print after this change is finished.
        """
        pass



import os
import subprocess
from pathlib import Path

import pytest

from mentat.diff_context import DiffContext
from mentat.errors import UserError

rel_path = Path("multifile_calculator/operations.py")


def _update_ops(temp_testbed, last_line, commit_message=None):
    # Update the last line of operations.py and (optionally) commit
    abs_path = os.path.join(temp_testbed, "multifile_calculator", "operations.py")
    with open(abs_path, "r") as f:
        lines = f.readlines()
    lines[-1:] = [
        f"    return {last_line}\n",
    ]
    with open(abs_path, "w") as f:
        f.writelines(lines)
    if commit_message:
        subprocess.run(["git", "add", abs_path], cwd=temp_testbed)
        subprocess.run(["git", "commit", "-m", commit_message], cwd=temp_testbed)


@pytest.fixture
def git_history(mock_config, temp_testbed):
    """Load a git repo with the following branches/commits:

    main
      'a / b' (from temp_testbed)
      'commit2'
      'commit3'
    test_branch (from commit2)
      'commit4'
    """
    _update_ops(temp_testbed, "commit2", "commit2")
    _update_ops(temp_testbed, "commit3", "commit3")
    subprocess.run(["git", "checkout", "HEAD~1"], cwd=temp_testbed)
    subprocess.run(["git", "checkout", "-b", "test_branch"], cwd=temp_testbed)
    # commit4
    _update_ops(temp_testbed, "commit4", "commit4")
    # Return on master commit3
    subprocess.run(["git", "checkout", "master"], cwd=temp_testbed)


def _get_file_message(temp_testbed):
    abs_path = os.path.join(temp_testbed, "multifile_calculator", "operations.py")
    file_message = ["/multifile_calculator/operations.py"]
    with open(abs_path, "r") as f:
        for i, line in enumerate(f.readlines()):
            file_message.append(f"{i}:{line}")
    return file_message


def test_diff_context_default(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    # DiffContext.__init__() (default): active code vs last commit
    diff_context = DiffContext()
    assert diff_context.target == "HEAD"
    assert diff_context.name == "HEAD (last commit)"
    assert diff_context.files == []

    # DiffContext.files (property): return git-tracked files with active changes
    _update_ops(temp_testbed, "commit5")
    diff_context._files_cache = None  # This is usually cached
    assert diff_context.files == [rel_path]

    # DiffContext.annotate_file_message(): modify file_message with diff
    file_message = _get_file_message(temp_testbed)
    annotated_message = diff_context.annotate_file_message(rel_path, file_message)
    expected = file_message[:-1] + [
        "14:-    return commit3",
        "14:+    return commit5",
    ]
    assert annotated_message == expected


def test_diff_context_commit(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    # Get the hash of 2-commits-ago
    last_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD~2"], cwd=temp_testbed, text=True
    ).strip()
    diff_context = DiffContext.create(diff=last_commit)
    assert diff_context.target == last_commit
    assert diff_context.name == f"{last_commit[:8]}: add testbed"
    assert diff_context.files == [rel_path]

    file_message = _get_file_message(temp_testbed)
    annotated_message = diff_context.annotate_file_message(rel_path, file_message)
    expected = file_message[:-1] + [
        "14:-    return a / b",
        "14:+    return commit3",
    ]
    assert annotated_message == expected


def test_diff_context_branch(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    diff_context = DiffContext.create(diff="test_branch")
    assert diff_context.target == "test_branch"
    assert diff_context.name.startswith("Branch test_branch:")
    assert diff_context.name.endswith(": commit4")
    assert diff_context.files == [rel_path]

    file_message = _get_file_message(temp_testbed)
    annotated_message = diff_context.annotate_file_message(rel_path, file_message)
    expected = file_message[:-1] + [
        "14:-    return commit4",
        "14:+    return commit3",
    ]
    assert annotated_message == expected


def test_diff_context_relative(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    diff_context = DiffContext.create(diff="HEAD~2")
    assert diff_context.target == "HEAD~2"
    assert diff_context.name.startswith("HEAD~2: ")
    assert diff_context.name.endswith(": add testbed")
    assert diff_context.files == [rel_path]

    file_message = _get_file_message(temp_testbed)
    annotated_message = diff_context.annotate_file_message(rel_path, file_message)
    expected = file_message[:-1] + [
        "14:-    return a / b",
        "14:+    return commit3",
    ]
    assert annotated_message == expected


def test_diff_context_errors(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    # Can't use both diff and pr_diff
    with pytest.raises(UserError) as e:
        DiffContext.create(diff="HEAD", pr_diff="master")
    assert str(e.value) == "Cannot specify more than one type of diff."

    # Invalid treeish
    with pytest.raises(UserError) as e:
        DiffContext.create(diff="invalid")
    assert str(e.value) == "Invalid treeish: invalid"


def test_diff_context_pr(
    mock_stream, mock_config, temp_testbed, git_history, mock_git_root
):
    subprocess.run(["git", "checkout", "test_branch"], cwd=temp_testbed)
    diff_context = DiffContext.create(pr_diff="master")

    commit2 = subprocess.check_output(
        ["git", "rev-parse", "HEAD~1"], cwd=temp_testbed, text=True
    ).strip()
    assert diff_context.target == commit2
    assert diff_context.name.startswith("Merge-base Branch master:")
    assert diff_context.name.endswith(": commit2")  # NOT the latest
    assert diff_context.files == [rel_path]



from enum import Enum
from pathlib import Path

from termcolor import colored
from typing_extensions import override

from mentat.code_file_manager import CodeFileManager
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    FileActionType,
    get_file_action_type,
)
from mentat.parsers.diff_utils import matching_index
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import Parser
from mentat.prompts.prompts import read_prompt

split_diff_parser_prompt_filename = Path("split_diff_parser_prompt.txt")


class SplitDiffDelimiters(Enum):
    FenceStart = "{fence[0]}"
    Start = "<<<<<<< HEAD"
    Middle = "======="
    End = ">>>>>>> updated"
    FenceEnd = "{fence[1]}"


class SplitDiffParser(Parser):
    @override
    def get_system_prompt(self) -> str:
        return read_prompt(split_diff_parser_prompt_filename)

    @override
    def provide_line_numbers(self) -> bool:
        return False

    @override
    def _code_line_beginning(
        self, display_information: DisplayInformation, cur_block: str
    ) -> str:
        lines = cur_block.split("\n")
        if SplitDiffDelimiters.Middle.value in lines:
            # Since we can't print surrounding lines for this parser,
            # no need for line_number_buffer - 1 spaces
            return colored("+", color="green")
        else:
            return colored("-", color="red")

    @override
    def _code_line_content(
        self,
        display_information: DisplayInformation,
        content: str,
        cur_line: str,
        cur_block: str,
    ) -> str:
        lines = cur_block.split("\n")
        if SplitDiffDelimiters.Middle.value in lines:
            return colored(content, color="green")
        else:
            return colored(content, color="red")

    @override
    def _could_be_special(self, cur_line: str) -> bool:
        line = cur_line.strip()
        # This stops us from printing the >>>>>> ====== and <<<<<< lines
        return any(
            delimiter.value.startswith(line) or line.startswith(delimiter.value)
            for delimiter in SplitDiffDelimiters
        )

    @override
    def _starts_special(self, line: str) -> bool:
        return line.strip().startswith(SplitDiffDelimiters.FenceStart.value)

    @override
    def _ends_special(self, line: str) -> bool:
        return line.strip().startswith(SplitDiffDelimiters.FenceStart.value)

    @override
    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        git_root: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        info = " ".join(special_block.strip().split(" ")[1:])
        is_creation = info.endswith(" +")
        is_deletion = info.endswith(" -")
        if is_creation or is_deletion:
            info = info[:-2]
        if " -> " in info:
            file_name, new_name = map(Path, info.split(" -> "))
        else:
            file_name, new_name = Path(info), None

        file_lines = self._get_file_lines(code_file_manager, rename_map, file_name)
        file_action_type = get_file_action_type(is_creation, is_deletion, new_name)
        display_information = DisplayInformation(
            file_name=file_name,
            file_lines=file_lines,
            added_block=[],
            removed_block=[],
            file_action_type=file_action_type,
            # Since we don't know where we're replacing until after the model outputs the removal block,
            # there's no way to print the surrounding lines
            first_changed_line=-1,
            last_changed_line=-1,
            new_name=new_name,
        )
        file_edit = FileEdit(
            git_root / file_name, [], is_creation, is_deletion, new_name
        )
        return (
            display_information,
            file_edit,
            file_action_type == FileActionType.UpdateFile,
        )

    @override
    def _ends_code(self, line: str) -> bool:
        return line.strip().startswith(SplitDiffDelimiters.FenceEnd.value)

    @override
    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ) -> str:
        # Find matching set of lines; first we check for a direct match,
        # then we check for a match excluding case, then we check for a match
        # excluding case and stripped. If we don't find one, we throw away this change.
        file_lines = self._get_file_lines(
            code_file_manager, rename_map, display_information.file_name
        )
        # Remove the delimiters, ending fence, and new line after ending fence
        lines = code_block.split("\n")[1:-3]
        middle_index = lines.index(SplitDiffDelimiters.Middle.value)
        removed_lines = lines[:middle_index]
        added_lines = lines[middle_index + 1 :]
        index = matching_index(file_lines, removed_lines)
        if index == -1:
            return colored(
                "Error: Original lines not found. Discarding this change.",
                color="red",
            )
        file_edit.replacements.append(
            Replacement(index, index + len(removed_lines), added_lines)
        )
        return ""



from __future__ import annotations

import logging
import os
import sys
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any, AsyncGenerator, Optional, cast

import openai
import openai.error
import tiktoken
from dotenv import load_dotenv
from openai.error import AuthenticationError

from mentat.session_stream import SESSION_STREAM

from .config_manager import mentat_dir_path
from .errors import MentatError, UserError

package_name = __name__.split(".")[0]

# openai doesn't seem to use type hints, so we have to use type: ignore and cast everywhere


# Check for .env file or already exported API key
# If no api key found, raise an error
def setup_api_key():
    if not load_dotenv(mentat_dir_path / ".env"):
        load_dotenv()
    key = os.getenv("OPENAI_API_KEY")
    base_url = os.getenv("OPENAI_API_BASE")
    if base_url:
        openai.api_base = base_url
    if not key:
        raise UserError(
            "No OpenAI api key detected.\nEither place your key into a .env"
            " file or export it as an environment variable."
        )
    try:
        openai.api_key = key
        openai.Model.list()  # type: ignore Test the API key
    except AuthenticationError as e:
        raise UserError(f"OpenAI gave an Authentication Error:\n{e}")


def is_test_environment():
    """Returns True if in pytest and not benchmarks"""
    benchmarks_running = os.getenv("MENTAT_BENCHMARKS_RUNNING")
    return (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and (not bool(benchmarks_running) or benchmarks_running == "false")
    )


async def _add_newline(response: AsyncGenerator[Any, None]):
    """
    The model normally ends it's response without a newline,
    but since our parsing relies on newlines to determine if a line is
    conversation or part of our format, adding a newline to the end
    makes everything much easier.
    """
    async for chunk in response:
        yield chunk
    yield {"choices": [{"delta": {"content": "\n"}}]}


async def call_llm_api(
    messages: list[dict[str, str]], model: str
) -> AsyncGenerator[Any, None]:
    if is_test_environment():
        logging.critical("OpenAI call attempted in non benchmark test environment!")
        raise MentatError("OpenAI call attempted in non benchmark test environment!")

    response: AsyncGenerator[Any, None] = cast(
        AsyncGenerator[Any, None],
        await openai.ChatCompletion.acreate(  # type: ignore
            model=model,
            messages=messages,
            temperature=0.5,
            stream=True,
        ),
    )

    return _add_newline(response)


# Ensures that each chunk will have at most one newline character
def chunk_to_lines(chunk: Any) -> list[str]:
    return chunk["choices"][0]["delta"].get("content", "").splitlines(keepends=True)


# NOTE: We may be calculating the length of Conversation messages incorrectly,
# but the difference should be negligible (<5 tokens per message):
# https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
def count_tokens(message: str, model: str) -> int:
    try:
        return len(
            tiktoken.encoding_for_model(model).encode(message, disallowed_special=())
        )
    except KeyError:
        return len(
            tiktoken.encoding_for_model("gpt-4").encode(message, disallowed_special=())
        )


def is_model_available(model: str) -> bool:
    available_models: list[str] = cast(
        list[str], [x["id"] for x in openai.Model.list()["data"]]  # type: ignore
    )

    return model in available_models


def model_context_size(model: str) -> Optional[int]:
    if "gpt-4" in model:
        if "32k" in model:
            return 32768
        else:
            return 8192
    elif "gpt-3.5" in model:
        if "16k" in model:
            return 16385
        else:
            return 4097
    else:
        return None


def model_price_per_1000_tokens(model: str) -> Optional[tuple[float, float]]:
    if "gpt-4" in model:
        if "32k" in model:
            return (0.06, 0.12)
        else:
            return (0.03, 0.06)
    elif "gpt-3.5" in model:
        if "16k" in model:
            return (0.003, 0.004)
        else:
            return (0.0015, 0.002)
    else:
        return None


async def get_prompt_token_count(messages: list[dict[str, str]], model: str) -> int:
    stream = SESSION_STREAM.get()

    prompt_token_count = 0
    for message in messages:
        prompt_token_count += count_tokens(message["content"], model)
    await stream.send(f"Total token count: {prompt_token_count}", color="cyan")

    token_buffer = 500
    context_size = model_context_size(model)
    if context_size:
        if prompt_token_count > context_size - token_buffer:
            await stream.send(
                f"Warning: {model} has a maximum context length of {context_size}"
                " tokens. Attempting to run anyway:",
                color="yellow",
            )
    return prompt_token_count


COST_TRACKER: ContextVar[CostTracker] = ContextVar("mentat:cost_tracker")


@dataclass
class CostTracker:
    total_tokens: int = 0
    total_cost: float = 0

    async def display_api_call_stats(
        self,
        num_prompt_tokens: int,
        num_sampled_tokens: int,
        model: str,
        call_time: float,
    ) -> None:
        stream = SESSION_STREAM.get()

        self.total_tokens += num_prompt_tokens + num_sampled_tokens
        tokens_per_second = num_sampled_tokens / call_time
        cost = model_price_per_1000_tokens(model)
        if cost:
            prompt_cost = (num_prompt_tokens / 1000) * cost[0]
            sampled_cost = (num_sampled_tokens / 1000) * cost[1]
            call_cost = prompt_cost + sampled_cost
            self.total_cost += call_cost

            speed_and_cost_string = (
                f"Speed: {tokens_per_second:.2f} tkns/s | Cost: ${call_cost:.2f}"
            )
        else:
            speed_and_cost_string = f"Speed: {tokens_per_second:.2f} tkns/s"
        await stream.send(speed_and_cost_string, color="cyan")

        costs_logger = logging.getLogger("costs")
        costs_logger.info(speed_and_cost_string)

    async def display_total_cost(self) -> None:
        stream = SESSION_STREAM.get()
        await stream.send(
            f"Total session cost: ${self.total_cost:.2f}", color="light_blue"
        )



from __future__ import annotations

import logging
import os
from contextvars import ContextVar
from pathlib import Path
from typing import TYPE_CHECKING

from mentat.git_handler import GIT_ROOT

# from .code_context import CODE_CONTEXT
from .errors import MentatError
from .session_input import ask_yes_no
from .session_stream import SESSION_STREAM
from .utils import sha256

if TYPE_CHECKING:
    # This normally will cause a circular import
    from .code_context import CodeContext
    from .parsers.file_edit import FileEdit

CODE_FILE_MANAGER: ContextVar[CodeFileManager] = ContextVar("mentat:code_file_manager")


class CodeFileManager:
    def __init__(self):
        self.file_lines = dict[Path, list[str]]()

    def read_file(self, path: Path) -> list[str]:
        git_root = GIT_ROOT.get()

        abs_path = path if path.is_absolute() else Path(git_root / path)
        rel_path = Path(os.path.relpath(abs_path, git_root))
        with open(abs_path, "r") as f:
            lines = f.read().split("\n")
        self.file_lines[rel_path] = lines
        return lines

    def _create_file(self, code_context: "CodeContext", abs_path: Path):
        logging.info(f"Creating new file {abs_path}")
        code_context.settings.paths.append(abs_path)
        # create any missing directories in the path
        abs_path.parent.mkdir(parents=True, exist_ok=True)
        with open(abs_path, "w") as f:
            f.write("")

    def _delete_file(self, code_context: "CodeContext", abs_path: Path):
        logging.info(f"Deleting file {abs_path}")
        if abs_path in code_context.include_files:
            del code_context.include_files[abs_path]
        abs_path.unlink()

    # Mainly does checks on if file is in context, file exists, file is unchanged, etc.
    async def write_changes_to_files(
        self,
        file_edits: list["FileEdit"],
        code_context: "CodeContext",
    ):
        stream = SESSION_STREAM.get()
        git_root = GIT_ROOT.get()

        for file_edit in file_edits:
            rel_path = Path(os.path.relpath(file_edit.file_path, git_root))
            if file_edit.is_creation:
                if file_edit.file_path.exists():
                    raise MentatError(
                        f"Model attempted to create file {file_edit.file_path} which"
                        " already exists"
                    )
                self._create_file(code_context, file_edit.file_path)
            else:
                if not file_edit.file_path.exists():
                    raise MentatError(
                        f"Attempted to edit non-existent file {file_edit.file_path}"
                    )
                elif file_edit.file_path not in code_context.include_files:
                    await stream.send(
                        f"Attempted to edit file {file_edit.file_path} not in context",
                        color="yellow",
                    )
                    continue

            if file_edit.is_deletion:
                await stream.send(
                    f"Are you sure you want to delete {rel_path}?", color="red"
                )
                if await ask_yes_no(default_yes=False):
                    await stream.send(f"Deleting {rel_path}...", color="red")
                    self._delete_file(code_context, file_edit.file_path)
                    continue
                else:
                    await stream.send(f"Not deleting {rel_path}", color="green")

            if not file_edit.is_creation:
                stored_lines = self.file_lines[rel_path]
                if stored_lines != self.read_file(rel_path):
                    logging.info(
                        f"File '{file_edit.file_path}' changed while generating changes"
                    )
                    await stream.send(
                        f"File '{rel_path}' changed while generating; current"
                        " file changes will be erased. Continue?",
                        color="light_yellow",
                    )
                    if not await ask_yes_no(default_yes=False):
                        await stream.send(f"Not applying changes to file {rel_path}")
                        continue
            else:
                stored_lines = []

            if file_edit.rename_file_path is not None:
                if file_edit.rename_file_path.exists():
                    raise MentatError(
                        f"Attempted to rename file {file_edit.file_path} to existing"
                        f" file {file_edit.rename_file_path}"
                    )
                self._create_file(code_context, file_edit.rename_file_path)
                self._delete_file(code_context, file_edit.file_path)
                file_edit.file_path = file_edit.rename_file_path

            new_lines = file_edit.get_updated_file_lines(stored_lines)
            with open(file_edit.file_path, "w") as f:
                f.write("\n".join(new_lines))

    def get_file_checksum(self, path: Path) -> str:
        if path.is_dir():
            return ""  # TODO: Build and maintain a hash tree for git_root
        return sha256(path.read_text())



import os
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.code_file_manager import CODE_FILE_MANAGER, CodeFileManager
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.session import Session


# Make sure we always give posix paths to GPT
@pytest.mark.asyncio
async def test_posix_paths(
    mock_stream, mock_config, mock_code_file_manager, mock_code_context, mock_parser
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = Path(os.path.join(dir_name, file_name))
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")

    mock_code_context.settings.paths = [file_path]
    mock_code_context._set_include_files()

    code_message = await mock_code_context.get_code_message(mock_config.model(), 1e6)
    assert dir_name + "/" + file_name in code_message.split("\n")


@pytest.mark.asyncio
async def test_partial_files(
    mock_stream, mock_config, mock_code_file_manager, mock_code_context, mock_parser
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = os.path.join(dir_name, file_name)
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write(dedent("""\
             I am a file
             with 5 lines
             third
             fourth
             fifth"""))

    file_path_partial = file_path + ":1,3-5"
    mock_code_context.settings.paths = [Path(file_path_partial)]
    mock_code_context.settings.auto_tokens = 0
    mock_code_context._set_include_files()
    mock_code_context.code_map = False

    code_message = await mock_code_context.get_code_message(
        mock_config.model(), max_tokens=1e6
    )
    assert code_message == dedent("""\
            Code Files:

            dir/file.txt
            1:I am a file
            3:third
            4:fourth
            5:fifth
              """)


@pytest.mark.asyncio
async def test_run_from_subdirectory(
    mock_collect_user_input, mock_call_llm_api, mock_setup_api_key
):
    """Run mentat from a subdirectory of the git root"""
    # Change to the subdirectory
    os.chdir("multifile_calculator")
    mock_collect_user_input.set_stream_messages(
        [
            (
                "Insert the comment # Hello on the first line of"
                " multifile_calculator/calculator.py and scripts/echo.py"
            ),
            "y",
            "q",
        ]
    )
    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment in both files.

        @@start
        {
            "file": "multifile_calculator/calculator.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end
        @@start
        {
            "file": "scripts/echo.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end""")])

    session = await Session.create(
        [Path("calculator.py"), Path("../scripts")], auto_tokens=0
    )
    await session.start()
    await session.stream.stop()

    # Check that it works
    with open("calculator.py") as f:
        calculator_output = f.readlines()
    with open("../scripts/echo.py") as f:
        echo_output = f.readlines()
    assert calculator_output[0].strip() == "# Hello"
    assert echo_output[0].strip() == "# Hello"


@pytest.mark.asyncio
async def test_changed_file(
    mocker,
    temp_testbed,
    mock_stream,
    mock_config,
    mock_collect_user_input,
    mock_code_context,
    mock_parser,
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = Path(temp_testbed) / dir_name / file_name
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")
    mock_code_context.settings.paths = [file_path]
    mock_code_context._set_include_files()

    # Load code_file_manager's file_lines
    code_file_manager = CodeFileManager()
    CODE_FILE_MANAGER.set(code_file_manager)
    code_file_manager.read_file(file_path)

    # Update the included files
    with open(file_path, "w") as file_file:
        file_file.write("I was a file")

    # Try to write_changes
    file_edit = FileEdit(
        file_path=file_path,
        replacements=[Replacement(0, 1, ["I am a file", "with edited lines"])],
    )
    assert await file_edit.is_valid()

    # Decline overwrite
    mock_collect_user_input.set_stream_messages(["n", "q"])
    await code_file_manager.write_changes_to_files([file_edit], mock_code_context)
    assert file_path.read_text().splitlines() == ["I was a file"]

    # Accept overwrite
    mock_collect_user_input.set_stream_messages(["y", "q"])
    await code_file_manager.write_changes_to_files([file_edit], mock_code_context)
    assert file_path.read_text().splitlines() == ["I am a file", "with edited lines"]



import json
import subprocess
import tempfile
from pathlib import Path
from typing import Any

from .session_stream import SESSION_STREAM
from .utils import run_subprocess_async


async def get_code_map(
    root: Path, file_path: Path, exclude_signatures: bool = False
) -> list[str]:
    # Create ctags from executable in a subprocess
    ctags_cmd_args = [
        "--extras=-F",
        "--input-encoding=utf-8",
        "--output-format=json",
        "--output-encoding=utf-8",
        # "--fields=+n",
    ]
    if exclude_signatures:
        ctags_cmd_args.append("--fields=-s")
    else:
        ctags_cmd_args.append("--fields=+S")
    ctags_cmd = ["ctags", *ctags_cmd_args, str(Path(root).joinpath(file_path))]

    output = await run_subprocess_async(*ctags_cmd)
    output_lines = output.splitlines()

    # Extract subprocess stdout into python objects
    ctags = set[tuple[Path, ...]]()
    for output_line in output_lines:
        try:
            tag = json.loads(output_line)
        except json.decoder.JSONDecodeError as err:
            await SESSION_STREAM.get().send(
                f"Error parsing ctags output: {err}\n{repr(output_line)}",
                color="yellow",
            )
            continue

        scope = tag.get("scope")
        kind = tag.get("kind")
        name = tag.get("name")
        signature = tag.get("signature")
        # line_number = tag.get("line")  # TODO: Use to split CodeFeatures in CodeContext

        last = name
        if signature:
            last += " " + signature

        res: list[Any] = []
        if scope:
            res.append(scope)
        res += [kind, last]

        ctags.add(tuple(res))

    if len(ctags) == 0:
        return []

    # Build LLM-readable string representation of ctag objects
    sorted_tags = sorted(ctags)
    output = ""
    last = [None] * len(sorted_tags[0])
    tab = "\t"
    for tag in sorted_tags:
        tag = list(tag)

        num_common = 0
        for i in range(len(last) + 1):
            if i == len(last):
                break
            if last[i] != tag[i]:
                num_common = i
                break

        indent = tab * num_common
        rest = tag[num_common:]

        for item in rest:
            output += indent + str(item) + "\n"
            indent += tab
        last = tag
    return output.splitlines()


async def check_ctags_disabled() -> str | None:
    try:
        cmd = ["ctags", "--version"]
        output = subprocess.check_output(cmd, stderr=subprocess.PIPE).decode("utf-8")
        output = output.lower()

        cmd = " ".join(cmd)
        if "universal ctags" not in output:
            return f"{cmd} does not claim to be universal ctags"
        if "+json" not in output:
            return f"{cmd} does not list +json support"

        with tempfile.TemporaryDirectory() as tempdir:
            hello_py = Path(tempdir) / "hello.py"
            with open(hello_py, "w", encoding="utf-8") as f:
                f.write("def hello():\n    print('Hello, world!')\n")
            await get_code_map(Path(tempdir), hello_py)
        return
    except FileNotFoundError:
        return "ctags executable not found"
    except Exception as e:
        return f"error running universal-ctags: {e}"



import math
import os
from enum import Enum
from pathlib import Path

from .code_file_manager import CODE_FILE_MANAGER
from .code_map import get_code_map
from .diff_context import annotate_file_message, parse_diff
from .git_handler import GIT_ROOT, get_diff_for_file
from .llm_api import count_tokens
from .parsers.parser import PARSER


class Interval:
    def __init__(
        self,
        start: int | float,
        end: int | float,
    ):
        self.start = start
        self.end = end

    def contains(self, line_number: int):
        return self.start <= line_number <= self.end


def parse_intervals(interval_string: str) -> list[Interval]:
    try:
        intervals = list[Interval]()
        for interval in interval_string.split(","):
            interval = interval.split("-", 1)
            if len(interval) == 1:
                intervals += [Interval(int(interval[0]), int(interval[0]))]
            else:
                intervals += [Interval(int(interval[0]), int(interval[1]))]
        return intervals
    except (ValueError, IndexError):
        return []


class CodeMessageLevel(Enum):
    CODE = ("code", 1, "Complete code")
    INTERVAL = ("interval", 2, "Specific range(s)")
    CMAP_FULL = ("cmap_full", 3, "Function/Class names and signatures")
    CMAP = ("cmap", 4, "Function/Class names")
    FILE_NAME = ("file_name", 5, "Relative path/filename")

    def __init__(self, key: str, rank: int, description: str):
        self.key = key
        self.rank = rank
        self.description = description


class CodeFile:
    """
    Represents a section of the code_message which is included with the prompt.
    Includes a section of code and an annotation method.

    Attributes:
        path: The absolute path to the file.
        intervals: The lines in the file.
        level: The level of information to include.
        diff: The diff annotations to include.
    """

    def __init__(
        self,
        path: str | Path,
        level: CodeMessageLevel = CodeMessageLevel.CODE,
        diff: str | None = None,
        user_included: bool = False,
    ):
        if Path(path).exists():
            self.path = Path(path)
            self.intervals = [Interval(0, math.inf)]
        else:
            path = str(path)
            split = path.rsplit(":", 1)
            self.path = Path(split[0])
            if not self.path.exists():
                self.path = Path(path)
                self.intervals = [Interval(0, math.inf)]
            else:
                self.intervals = parse_intervals(split[1])
                level = CodeMessageLevel.INTERVAL
        self.level = level
        self.diff = diff
        self.user_included = user_included

    def __repr__(self):
        return (
            f"CodeFile(fname={self.path.name}, intervals={self.intervals},"
            f" level={self.level}, diff={self.diff})"
        )

    def contains_line(self, line_number: int):
        return any([interval.contains(line_number) for interval in self.intervals])

    async def _get_code_message(self) -> list[str]:
        git_root = GIT_ROOT.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        parser = PARSER.get()

        code_message: list[str] = []

        # We always want to give GPT posix paths
        abs_path = Path(git_root / self.path)
        rel_path = Path(os.path.relpath(abs_path, git_root))
        posix_rel_path = Path(rel_path).as_posix()
        if self.user_included:
            filename = f"USER INCLUDED: {posix_rel_path}"
        else:
            filename = f"{posix_rel_path}"
        code_message.append(filename)

        if self.level in {CodeMessageLevel.CODE, CodeMessageLevel.INTERVAL}:
            file_lines = code_file_manager.read_file(abs_path)
            for i, line in enumerate(file_lines, start=1):
                if self.contains_line(i):
                    if parser.provide_line_numbers():
                        code_message.append(f"{i}:{line}")
                    else:
                        code_message.append(f"{line}")
        elif self.level == CodeMessageLevel.CMAP_FULL:
            cmap = await get_code_map(git_root, self.path)
            code_message += cmap
        elif self.level == CodeMessageLevel.CMAP:
            cmap = await get_code_map(git_root, self.path, exclude_signatures=True)
            code_message += cmap
        code_message.append("")

        if self.diff is not None:
            diff: str = get_diff_for_file(self.diff, rel_path)
            diff_annotations = parse_diff(diff)
            if self.level == CodeMessageLevel.CODE:
                code_message = annotate_file_message(code_message, diff_annotations)
            else:
                for section in diff_annotations:
                    code_message += section.message
        return code_message

    _file_checksum: str | None = None
    _code_message: list[str] | None = None

    async def get_code_message(self) -> list[str]:
        git_root = GIT_ROOT.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        abs_path = git_root / self.path
        file_checksum = code_file_manager.get_file_checksum(Path(abs_path))
        if file_checksum != self._file_checksum or self._code_message is None:
            self._file_checksum = file_checksum
            self._code_message = await self._get_code_message()
        return self._code_message

    async def count_tokens(self, model: str) -> int:
        code_message = await self.get_code_message()
        return count_tokens("\n".join(code_message), model)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

from mentat.session_stream import SESSION_STREAM

from .code_context import CODE_CONTEXT
from .code_file import CodeFile
from .errors import MentatError
from .git_handler import commit


class Command(ABC):
    # Unfortunately, Command isn't defined here yet, so even with annotations we need quotation marks
    _registered_commands = dict[str, type["Command"]]()

    def __init_subclass__(cls, command_name: str | None) -> None:
        if command_name is not None:
            Command._registered_commands[command_name] = cls

    @classmethod
    def create_command(cls, command_name: str) -> Command:
        if command_name not in cls._registered_commands:
            return InvalidCommand(command_name)

        command_cls = cls._registered_commands[command_name]
        return command_cls()

    @classmethod
    def get_command_completions(cls) -> List[str]:
        return list(map(lambda name: "/" + name, cls._registered_commands))

    @abstractmethod
    async def apply(self, *args: str) -> None:
        pass

    # TODO: make more robust way to specify arguments for commands
    @classmethod
    @abstractmethod
    def argument_names(cls) -> list[str]:
        pass

    @classmethod
    @abstractmethod
    def help_message(cls) -> str:
        pass


class InvalidCommand(Command, command_name=None):
    def __init__(self, invalid_name: str):
        self.invalid_name = invalid_name

    async def apply(self, *args: str) -> None:
        await SESSION_STREAM.get().send(
            f"{self.invalid_name} is not a valid command. Use /help to see a list of"
            " all valid commands",
            color="light_yellow",
        )

    @classmethod
    def argument_names(cls) -> list[str]:
        raise MentatError("Argument names called on invalid command")

    @classmethod
    def help_message(cls) -> str:
        raise MentatError("Help message called on invalid command")


help_message_width = 60


class HelpCommand(Command, command_name="help"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()

        if not args:
            commands = Command._registered_commands.keys()
        else:
            commands = args
        for command_name in commands:
            if command_name not in Command._registered_commands:
                await stream.send(
                    f"Error: Command {command_name} does not exist.", color="red"
                )
            else:
                command_class = Command._registered_commands[command_name]
                argument_names = command_class.argument_names()
                help_message = command_class.help_message()
                message = (
                    " ".join(
                        [f"/{command_name}"] + [f"<{arg}>" for arg in argument_names]
                    ).ljust(help_message_width)
                    + help_message
                )
                await stream.send(message)

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Displays this message"


class CommitCommand(Command, command_name="commit"):
    default_message = "Automatic commit"

    async def apply(self, *args: str) -> None:
        if args:
            commit(args[0])
        else:
            commit(self.__class__.default_message)

    @classmethod
    def argument_names(cls) -> list[str]:
        return [f"commit_message={cls.default_message}"]

    @classmethod
    def help_message(cls) -> str:
        return "Commits all of your unstaged and staged changes to git"


class IncludeCommand(Command, command_name="include"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()

        if len(args) == 0:
            await stream.send("No files specified\n", color="yellow")
            return
        for file_path in args:
            code_file = CodeFile(file_path)
            await code_context.include_file(code_file)

    @classmethod
    def argument_names(cls) -> list[str]:
        return ["file1", "file2", "..."]

    @classmethod
    def help_message(cls) -> str:
        return "Add files to the code context"


class ExcludeCommand(Command, command_name="exclude"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()

        if len(args) == 0:
            await stream.send("No files specified\n", color="yellow")
            return
        for file_path in args:
            code_file = CodeFile(file_path)
            await code_context.exclude_file(code_file)

    @classmethod
    def argument_names(cls) -> list[str]:
        return ["file1", "file2", "..."]

    @classmethod
    def help_message(cls) -> str:
        return "Remove files from the code context"
